/* Open Digita Services  --  Camera Device Protocol definitions.

   Copyright (C) 1998 James C. Thompson

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc., 59
   Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/*--------------------------------------------------------------------------
 * System include files
 */ 

#include <fcntl.h>
#include <termio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <memory.h>

/*--------------------------------------------------------------------------
 * Local include files
 */ 

#include "cameraP.h"
#include "camera.h"
#include "cdpP.h"
#include "misc.h"
#include "wire.h"
#include "poll-mode-send.h"
#include "poll-mode-rec.h"

/* These are declared in system header files, but for some reason, the
   strict ANSI checking in gcc doesn't seem to acknowledge those
   declarations. */
extern void cfmakeraw(struct termios *);
extern void usleep(unsigned int);

/* Start Editing Here: */

ODSCamera odsCreateCamera()
{
  /* Allocate a camera */
  ODSCamera camera = (ODSCamera) malloc(sizeof(ODSCameraRec));

  if (camera == 0)
    {
      return camera;
    }

  /* Initialize its fields to default values. */
  camera->type = ods_unknown;
  camera->mode = ods_closed;
  camera->fd = -1;
  camera->buffer = 0;
  camera->data_speed = 9600;
  camera->timeout = 10000;
  camera->error = kODSNoErr;

  return camera;
}

ODSResult odsDestroyCamera(ODSCamera camera)
{
  /* Release the memory we've used. */
  free(camera);

  return kODSNoErr;
}

ODSResult odsOpenNSS(ODSCamera camera, const char *device,
		     unsigned long req_speed)
{
  ODSBeacon		beacon;
  ODSBeaconAck		beacon_ack;
  ODSBeaconComp		beacon_comp;
  speed_t		speed;

  struct termios tty;

  if ((camera == 0) || (device == 0))
    {
      return kODSParamError;
    }

  if ((camera->mode != ods_closed) || (camera->fd >= 0))
    {
      return camera->error = kODSAlreadyOpen;	/* Assignment */
    }

  camera->fd = open(device, O_RDWR|O_NOCTTY|O_SYNC);

  /* Keep the data speed at least 9600, because the camera doesn't
     communicate well at slower speeds.  We have to be able to support 
     at least 9600 anyway, to complete the beacon phase, so it should
     be safe. */

  if (req_speed < 9600)
    {
      req_speed = 9600;
    }

  /* Open the port */
  if (camera->fd < 0)
    {
      return camera->error = kODSUnixError; /* Assignment */
    }

  camera->mode = ods_presend;

  /* Make a copy of the tty settings to restore later. */
  tcgetattr(camera->fd, &(camera->tty_original));

  /* Put it in raw mode */
  memset(&tty, 0, sizeof(tty));
  cfmakeraw(&tty);

  /* ----- Zeroth phase: toggle the DTR by setting the rate to zero (turning
     off the port) then setting the rate back to 9600.  This will initiate
     beacon phase from the camera. */

  /* off... */
  cfsetospeed(&tty, B0);
  cfsetispeed(&tty, B0);
  tcsetattr(camera->fd, TCSANOW, &tty);

  /* ...wait... */
  usleep(50);

  /* ...on. */
  cfsetospeed(&tty, B9600);
  cfsetispeed(&tty, B9600);
  tcsetattr(camera->fd, TCSANOW, &tty);

  /* ----- First phase: read the beacon from the camera */

#if defined(DEBUG)
  printf("----- \n");
#endif

  if (!read_beacon(camera->fd, &beacon))
    {
      close(camera->fd);
      camera->fd = -1;
      camera->mode = ods_closed;
      return camera->error = kODSIOError; /* Assignment */
    }

  swap_beacon(&beacon);

#if defined(DEBUG)
  dump_beacon(&beacon);
#endif
    
  if (!check_beacon(&beacon))
    {
      close(camera->fd);
      camera->fd = -1;
      camera->mode = ods_closed;
      return camera->error = kODSIOError; /* Assignment */
    }

#if defined(DEBUG)
  printf("\n----- \n");
#endif

  /* ----- Second phase: send the beacon acknowledgment to the camera */

  beacon_ack.intro = 0x5aa5;
  beacon_ack.interface_type = 0x55;
  beacon_ack.cf_reserved = 0;
  beacon_ack.cf_pod_receive_mode = 0;
  beacon_ack.cf_host_receive_mode = 0;
  beacon_ack.data_speed = req_speed;
  beacon_ack.device_frame_size = 1023;
  beacon_ack.host_frame_size = 1023;

  swap_beacon_ack(&beacon_ack);
  prepare_beacon_ack(&beacon_ack);
    
  if (!write_beacon_ack(camera->fd, &beacon_ack))
    {
      close(camera->fd);
      camera->fd = -1;
      camera->mode = ods_closed;
      return camera->error = kODSIOError; /* Assignment */
    }

#if defined(DEBUG)
  swap_beacon_ack(&beacon_ack);
  dump_beacon_ack(&beacon_ack);
#endif

#if defined(DEBUG)
  printf("\n----- \n");
#endif

  /* ----- Third phase: read the beacon completion from the camera */

  usleep(20000);

  if (!read_beacon_comp(camera->fd, &beacon_comp))
    {
      close(camera->fd);
      camera->fd = -1;
      camera->mode = ods_closed;
      return camera->error = kODSIOError; /* Assignment */
    }

  swap_beacon_comp(&beacon_comp);

  /* Pick up the rate and buffer sizes from the completion message. */
  camera->data_speed = beacon_comp.data_speed;
  camera->host_size = beacon_comp.host_frame_size;
  camera->device_size = beacon_comp.device_frame_size;

#if defined(DEBUG)
  dump_beacon_comp(&beacon_comp);
#endif

  if (beacon_comp.result != 0)
    {
      close(camera->fd);
      camera->fd = -1;
      camera->mode = ods_closed;
      return camera->error = kODSParamError; /* Assignment */
    }

  /* Pause T1 ms */
  usleep(100000);

  /* Change baud rate here */
  if (camera->data_speed >= 115200)
    {
      speed = B115200;
    }
  else if (camera->data_speed >= 57600)
    {
      speed = B57600;
    }
  else if (camera->data_speed >= 38400)
    {
      speed = B38400;
    }
  else if (camera->data_speed >= 19200)
    {
      speed = B19200;
    }
  else
    {
      speed = B9600;
    }

  cfsetospeed(&tty, speed);
  cfsetispeed(&tty, speed);
  tcsetattr(camera->fd, TCSANOW, &tty);

  /* Pause T1 ms */
  usleep(100000);

  return kODSNoErr;
}

ODSResult odsCloseNSS(ODSCamera camera)
{
  /* Make sure the port's open */
  if ((camera->mode == ods_closed) || (camera->fd < 0))
    {
      return camera->error = kODSNotOpen; /* Assignment */
    }

  /* Restore its TTY settings */
  tcsetattr(camera->fd, TCSANOW, &(camera->tty_original));

  /* Close the port and mark it as closed. */
  close(camera->fd);
  camera->fd = -1;

  return kODSNoErr;
}

/*--------------------------------------------------------------------------
 * Product and Image Information Commands
 */ 

ODSResult odsGetProductInfo(ODSCamera camera, PName pname, 
			    ResLength *reslength,
			    PNameTypeValueStruct *valuelist,
			    int valuelen)
{
  /* The message length depends on whether a pname was specified */
  unsigned long send_length = (pname == 0) ? 0 : 4;

  /**** Send phase ****/

  begin_send(camera, send_length, ods_get_product_info);

  /* Begin sending command-specific parameters. */
  if (pname != 0)
    {
      send_uint(camera, pname);
    }

  end_send(camera);

  usleep(2000);

  /**** Receive phase ****/

  /* Begin receiving command-specific results. */
  {
    int i;
    unsigned long length;

    begin_rec(camera, &length);
    rec_uint(camera, reslength);
    
    for (i = 0; i < *reslength; i++)
      {
	PNameTypeValueStruct ptvs;

	read_ptvs(camera, &ptvs);

	if (i < valuelen)
	  {
	    valuelist[i] = ptvs;
	  }
      }
  }
  
  end_rec(camera);

  return kODSNoErr;
}

ODSResult odsSetProductInfo(ODSCamera camera, UInteger reqlength,
			    PNameTypeValueStruct *list,
			    UInteger *driveno, String *path,
			    SInteger *availablespace)
{
  return kODSUnimplemented;
}

ODSResult odsGetImageSpecifications(ODSCamera camera, 
				    UInteger *imagespecificationlist)
{
  return kODSUnimplemented;
}

/*--------------------------------------------------------------------------
 * Status Commands
 */ 

ODSResult odsGetCameraStatus(ODSCamera camera, BitFlags *systemstatus,
			     BitFlags *capturestatus, BitFlags
			     *vendorstatus)
{
  return kODSUnimplemented;
}

ODSResult odsGetError(ODSCamera camera, ErrorData *errordata)
{
  return kODSUnimplemented;
}

/*--------------------------------------------------------------------------
 * Camera Capabilities and State Commands
 */ 

ODSResult odsGetCameraCapabilities(ODSCamera camera, PName pname,
				   UInteger *reslength,
				   CapabilitiesData *capabilitiesdata,
				   unsigned long valuelen)
{
  return kODSUnimplemented;
}

ODSResult odsGetCameraState(ODSCamera camera, PName pname,
			    UInteger *resleng,
			    PNameTypeValueStruct *typevaluelist,
			    unsigned long valuelen)
{
  return kODSUnimplemented;
}

ODSResult odsSetCameraState(ODSCamera camera,
			    const PNameValueStruct *valuestruct)
{
  return kODSUnimplemented;
}

ODSResult odsGetCameraDefault(ODSCamera camera, UInteger defaultsource,
			      PName pname, UInteger *reslength,
			      PNameTypeValueStruct *typevaluelist,
			      unsigned long valuelen)
{
  return kODSUnimplemented;
}

ODSResult odsSetCameraDefault(ODSCamera camera, UInteger updatesource,
			      PNameValueStruct *valuestruct)
{
  return kODSUnimplemented;
}

ODSResult odsRestoreCameraState(ODSCamera camera, UInteger defaultsource,
				PName pname)
{
  return kODSUnimplemented;
}

/*--------------------------------------------------------------------------
 * Scene Analysis Command
 */ 

ODSResult odsGetSceneAnalysis(ODSCamera camera, UInteger analysistype,
			      Boolean preview, UInteger *reslength,
			      UInteger *sceneanalysisdata)
{
  return kODSUnimplemented;
}

/*--------------------------------------------------------------------------
 * Power and Capture Commands
 */ 

ODSResult odsGetPowerMode(ODSCamera camera, UInteger *PowerState)
{
  unsigned long length;

  /**** Send phase ****/

  begin_send(camera, 0, ods_get_power_mode);
  end_send(camera);

  usleep(2000);

  /**** Receive phase ****/

  begin_rec(camera, &length);
  rec_uint(camera, PowerState);
  
  end_rec(camera);

  return kODSNoErr;
}

ODSResult odsSetPowerMode(ODSCamera camera)
{
  unsigned long length;

  /**** Send phase ****/

  begin_send(camera, 0, ods_set_power_mode);
  end_send(camera);

  usleep(2000);

  /**** Receive phase ****/

  begin_rec(camera, &length);
  end_rec(camera);

  return kODSNoErr;
}

ODSResult odsGetS1Mode(ODSCamera camera, Boolean *s1mode)
{
  return kODSUnimplemented;
}

ODSResult odsSetS1Mode(ODSCamera camera, Boolean s1mode)
{
  return kODSUnimplemented;
}

ODSResult odsStartCapture(ODSCamera camera)
{
  return kODSUnimplemented;
}

/*--------------------------------------------------------------------------
 * File Commands
 */ 

ODSResult odsGetFileList(ODSCamera camera, UInteger listorder,
			 const FileNameStruct *filename,
			 UInteger *reslength, ResFileItem *resfilelist,
			 int valuelen)
{
  unsigned long length;
  int i;

  /* The message length depends on whether a filename was specified */
  unsigned long send_length = (filename == 0) ? 4 : 56;

  /**** Send phase ****/

  begin_send(camera, send_length, ods_get_file_list);  /* Begin with command */
  send_uint(camera, listorder);		    /* Send list order */

  if (filename != 0)
    {
      /* Send the filename if there is one. */
      write_filename(camera, filename);
    }

  end_send(camera);			    /* done sending */

  usleep(2000);

  /**** Receive phase ****/

  begin_rec(camera, &length);		    /* Begin receiving */
  rec_uint(camera, reslength);		    /* Get result length */

  /* Loop over file items in result-list. */
  for (i = 0; i < *reslength; i++)
    {
      ResFileItem file;

      /* Receive the file item */
      read_fileitem(camera, &file);

      /* If there's room in the return list, save the file item. */
      if (i < valuelen)
	{
	  resfilelist[i] = file;
	}
    }

  end_rec(camera);			    /* Done receiving */

  return kODSNoErr;
}

ODSResult odsGetNewFileList(ODSCamera camera, UInteger *reslength, 
			    ResFileItem *resfilelist, int valuelen)
{
  return kODSUnimplemented;
}

ODSResult odsGetFileData(ODSCamera camera, const FileNameStruct *filename,
			 UInteger dataselector, PartialTag *partialtag, 
			 void *filedata)
{
  unsigned long length;			    /* Length of message */
  unsigned long available_length;	    /* Number of bytes
					       available to download. */
  unsigned long dllength;		    /* Number of bytes downloaded */

  dllength = partialtag->Length;

  /**** Send phase ****/

  begin_send(camera, 68, ods_get_file_data);/* Send command */
  write_filename(camera, filename);	    /* FilenameStruct */
  send_uint(camera, dataselector);	    /* DataSelector */
  write_partialtag(camera, partialtag);	    /* PartialTag */
  end_send(camera);			    /* Done */

  usleep(2000);

  /**** Receive phase ****/

  begin_rec(camera, &length);
  read_partialtag(camera, partialtag);

  /* Figure out the number of bytes available for download in this
     partial.  The available number changes, depending on whether
     we're downloading a thumbnail, because thumbnails send 4 unsigned 
     ints *in addition to* the number of bytes indicated in 'length'. */
  if (dataselector == 0)
    {
      available_length = partialtag->Length;
    }
  else
    {
      available_length = 16 + (partialtag->Length);
    }

  /* If we were expecting to download more data than was available,
     then reduce the download length to the number available. */
  if (dllength > available_length)
    {
      dllength = available_length;
    }

  rec_bytes(camera, filedata, dllength);
  end_rec(camera);
  
  return kODSNoErr;
}

ODSResult odsSetFileData(ODSCamera camera, const FileNameStruct *filename,
			 UInteger dataselector, const PartialTag *partialtag,
			 const void *filedata)
{
  return kODSUnimplemented;
}

ODSResult odsEraseFile(ODSCamera camera, const FileNameStruct *filename)
{
  return kODSUnimplemented;
}

ODSResult odsGetStorageStatus(ODSCamera camera, UInteger *takencount,
			      UInteger *availablecount, SInteger rawcount)
{
  return kODSUnimplemented;
}

ODSResult odsGetFileTag(ODSCamera camera, 
			const FileNameStruct *name,
			PName pname,
			ResLength *reslength,
			PNameTypeValueStruct *valuelist,
			int valuelen)
{
  /* The message length depends on whether a pname was specified */
  unsigned long send_length = 56; /*(pname == 0) ? 52 : 56;*/

  /**** Send phase ****/
  begin_send(camera, send_length, ods_get_file_tag);
  write_filename(camera, name);	    /* FilenameStruct */
  send_uint(camera, pname);
  end_send(camera);

  usleep(2000);

  /**** Receive phase ****/

  /* Begin receiving command-specific results. */
  {
    int i;
    unsigned long length;

    begin_rec(camera, &length);

    /* Receive the number of PNameTypeValueStructs in the reply. */
    rec_uint(camera, reslength);
    
    /* Loop over each of the structs, reading each. */
    for (i = 0; i < *reslength; i++)
      {
	PNameTypeValueStruct ptvs;

	/* Read the reply struct */
	read_ptvs(camera, &ptvs);

	/* If there's room in the caller's reply list, store the struct */
	if (i < valuelen)
	  {
	    valuelist[i] = ptvs;
	  }
      }
  }
  
  end_rec(camera);

  return kODSNoErr;
}

ODSResult odsSetUserFileTag(ODSCamera camera, const FileNameStruct *filename,
			    const PNameValueStruct *valuestruct)
{
  return kODSUnimplemented;
}

/*--------------------------------------------------------------------------
 * Clock Commands
 */ 

ODSResult odsSetClock(ODSCamera camera, const ClockDateTime *datetime)
{
  return kODSUnimplemented;
}

ODSResult odsGetClock(ODSCamera camera, ClockDateTime *datetime)
{
  return kODSUnimplemented;
}

ODSResult odsGetInterfaceTimeout(ODSCamera camera, Timeout *timeout)
{
  return kODSUnimplemented;
}

ODSResult odsSetInterfaceTimeout(ODSCamera camera, const Timeout *timeout)
{
  return kODSUnimplemented;
}
