/* Open Digita Services  --  Camera Device Protocol definitions.

   Copyright (C) 1998 James C. Thompson

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc., 59
   Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/*--------------------------------------------------------------------------
 * System include files
 */ 

#include <unistd.h>
#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*--------------------------------------------------------------------------
 * Local include files
 */ 

#include "ods.h"
#include "cameraP.h"

/* Start Editing Here: */

typedef struct _camopts
{
  unsigned int		opt_type : 1;   /* Device type specified */
  unsigned int		opt_serial : 1;	/* Prefer NSS */
  unsigned int		opt_usb : 1;	/* Prefer Useless Serial Bus */
  unsigned int		opt_irda : 1;	/* Prefer infrared */

  const char *		devtype;        /* Type of device to use */
  const char *		device;	        /* Device to use. */
  const char *		bitrate;	/* Bitrate for serial devs */

  unsigned int		rate;		/* Numeric bitrate */

} camopts;

static const char *strcdup(const char *orig, const char *src)
{
  if (orig != 0)
    {
      free((char *) orig);
    }

  if (src != 0)
    {
      return strdup(src);
    }

  return 0;
}

static void _ods_deduce_devtype(ODSCamera camera)
{
  camopts *opts = camera->opts;

  if ((strncmp(opts->device, "/dev/cua", 8) == 0) ||
      (strncmp(opts->device, "/dev/tty", 8) == 0))
    {
#if DEBUG
      printf("deducing serial\n");
#endif
      opts->opt_serial = !0;
      opts->opt_usb = 0;
      opts->opt_irda = 0;
    }
  else if (strncmp(opts->device, "/dev/usb", 8) == 0)
    {
#if DEBUG
   printf("deducing usb\n");
#endif
      opts->opt_serial = 0;
      opts->opt_usb = !0;
      opts->opt_irda = 0;
    }
  else if (strncmp(opts->device, "/dev/irda", 9) == 0)
    {
#if DEBUG
   printf("deducing irda\n");
#endif
      opts->opt_serial = 0;
      opts->opt_usb = 0;
      opts->opt_irda = !0;
    }
}

static void _ods_get_devtype(ODSCamera camera)
{
  camopts *opts = camera->opts;

  if (strncmp(opts->devtype, "serial", strlen(opts->devtype)) == 0)
    {
#if DEBUG
      printf("getting serial\n");
#endif
      opts->opt_serial = !0;
      opts->opt_usb = 0;
      opts->opt_irda = 0;
    }
  else if (strncmp(opts->devtype, "usb", strlen(opts->devtype)) == 0)
    {
#if DEBUG
      printf("getting usb\n");
#endif
      opts->opt_serial = 0;
      opts->opt_usb = !0;
      opts->opt_irda = 0;
    }
  else if (strncmp(opts->devtype, "irda", strlen(opts->devtype)) == 0)
    {
#if DEBUG
      printf("getting irda\n");
#endif
      opts->opt_serial = 0;
      opts->opt_usb = 0;
      opts->opt_irda = !0;
    }
}

static void _ods_get_bitrate(ODSCamera camera)
{
  camopts *opts = camera->opts;
  unsigned long rate = atol(opts->bitrate);
  
  if (rate != 0)
    {
      opts->rate = rate;
    }
}

static void _ods_init_env(ODSCamera camera)
{
  camopts *opts = camera->opts;

  const char *devtype = getenv("ODSTYPE");
  const char *device = getenv("ODSDEVICE");
  const char *bitrate = getenv("ODSRATE");

  if (devtype != 0)
    {
      opts->devtype = strcdup(opts->devtype, devtype);
      opts->opt_type = !0;
      _ods_get_devtype(camera);
    }
  
  if (device != 0)
    {
      opts->device = strcdup(opts->device, device);
      _ods_deduce_devtype(camera);
    }

  if (bitrate != 0)
    {
      opts->bitrate = strcdup(opts->bitrate, bitrate);
      _ods_get_bitrate(camera);
    }
}

static void _ods_init_defaults(ODSCamera camera)
{
  camopts *opts = camera->opts;

  opts->opt_serial = !0;
  opts->device = strcdup(opts->device, "/dev/ttyS0");
  opts->rate = 9600;
}

static void _ods_help(const char *argv[])
{
  printf("Usage: %s [OPTION]... [COMMAND]...\n", argv[0]);
  printf("Connect to camera and perform COMMANDs.\n\n");

  printf("Development note: currently COMMANDs are unsupported; default\n");
  printf("action is to copy all photo files from camera.  However, all listed\n");
  printf("OPTIONS are now supported.\n\n");

  printf("  -d DEV, --device=DEV       Look for camera at device DEV.\n");
  printf("  -r RATE, --rate=RATE       Use bit-rate RATE for serial device\n");
  printf("  -t TYPE, --type=TYPE       Specifies device type, one of:\n");
  printf("                             \"serial\", \"usb\", or \"irda\"\n");
  printf("  -i, --irda                 Same as --device-type=irda\n");
  printf("  -s, --serial               Same as --device-type=serial\n");
  printf("  -u, --usb                  Same as --device-type=usb\n");
  printf("  -h, --help                 Display this help and exit.\n");

  printf("\nReport bugs to jim.thompson@pobox.com\n");

  exit(0);
}

static void _ods_init_args(ODSCamera camera, int *argc, const char *argv[])
{
  camopts *opts = camera->opts;

  int i;
  int c;

  opterr = 0;			/* Tell getopt not to report errors. */

  while (1)
    {
      int startind = optind ? optind : 1;
      int dont_mangle = 0;
      int option_index = 0;

      /* For each camera-specific --option, we also recognize
         --ods-option.  Does not apply to generic options such as
	 --help. */

      static struct option long_options[] =
      {
	{ "device", 1, 0, 'd' },
	{ "ods-device", 1, 0, 'd' },

	{ "rate", 1, 0, 'r' },
	{ "ods-rate", 1, 0, 'r' },

	{ "type", 1, 0, 't' },
	{ "ods-type", 1, 0, 't' },

	{ "serial", 0, 0, 's' },
	{ "ods-serial", 0, 0, 's' },

	{ "usb", 0, 0, 'u' },
	{ "ods-usb", 0, 0, 'u' },

	{ "irda", 0, 0, 'i' },
	{ "ods-irda", 0, 0, 'i' },

	{ "help", 0, 0, 'h' },

	{0, 0, 0, 0}
      };

      c = getopt_long (*argc, (char *const *) argv, "-d:r:t:suih",
		       long_options, &option_index);

      if (c == -1)
	break;

      switch (c)
	{
	case 1:
	  dont_mangle = !0;
	  break;

	case 'd':
#if DEBUG
	  printf ("device option with value `%s'\n", optarg);
#endif

	  opts->device = strcdup(opts->device, optarg);

	  /* If no device type has already been specified, attempt to
	     guess the device type based on the device name. */
	  if (!(opts->opt_type))
	    {
	      _ods_deduce_devtype(camera);
	    }

	  break;

	case 't':
#if DEBUG
	  printf ("device-type option with value `%s'\n", optarg);
#endif

	  opts->opt_type = !0;
	  opts->devtype = strcdup(opts->devtype, optarg);
	  _ods_get_devtype(camera);

	  break;

	case 'r':
#if DEBUG
	  printf ("serial-rate option with value `%s'\n", optarg);
#endif

	  opts->bitrate = strcdup(opts->bitrate, optarg);
	  _ods_get_bitrate(camera);

	  break;

	case 's':
#if DEBUG
	  printf ("use-serial option\n");
#endif
	  opts->opt_type = !0;
	  opts->opt_serial = !0;
	  opts->opt_usb = 0;
	  opts->opt_irda = 0;
	  break;

	case 'u':
#if DEBUG
	  printf ("use-usb option\n");
#endif
	  opts->opt_type = !0;
	  opts->opt_serial = 0;
	  opts->opt_usb = !0;
	  opts->opt_irda = 0;
	  break;

	case 'i':
#if DEBUG
	  printf ("use-irda option\n");
#endif
	  opts->opt_type = !0;
	  opts->opt_serial = 0;
	  opts->opt_usb = 0;
	  opts->opt_irda = !0;
	  break;

	case 'h':
	  _ods_help(argv);
	  break;

	case '?':
	  dont_mangle = !0;
	  break;

	default:
	  dont_mangle = !0;
	}

      if (!dont_mangle)
	{
	  int diff = optind - startind;
	  
	  for (i = optind; i < *argc; i++)
	    {
	      argv[i - diff] = argv[i];
	    }
	  
	  optind -= diff;
	  (*argc) -= diff;
	}
    }

#ifdef DEBUG
  for (i = 0; i < *argc; i++)
    {
      printf("argv[%d] = <%s>\n", i, argv[i]);
    }
#endif
}

#if DEBUG
static void _dump_opts(ODSCamera camera)
{
  camopts *opts = camera->opts;

  if (camera->opts == 0)
    {
      return;
    }

  printf("------------------------\n");

  if (opts->device == 0)
    {
      printf("no device specified\n");
    }
  else
    {
      printf("device = <%s>\n", opts->device);
    }

  if (opts->devtype == 0)
    {
      printf("no devtype specified\n");
    }
  else
    {
      printf("devtype = <%s>\n", opts->devtype);
    }

  if (opts->devtype == 0)
    {
      printf("no bitrate specified\n");
    }
  else
    {
      printf("bitrate = <%s>\n", opts->bitrate);
    }

  printf("opt_devtype = %d\n", opts->opt_type);
  printf("opt_serial = %d\n", opts->opt_serial);
  printf("opt_usb = %d\n", opts->opt_usb);
  printf("opt_irda = %d\n", opts->opt_irda);

  printf("rate = %d\n", opts->rate);

  printf("------------------------\n");
}
#endif

ODSCamera ODSInitialize(int *argc, const char *argv[])
{
  ODSCamera camera;
  camopts *opts;

  camera = odsCreateCamera();

  camera->opts = opts = malloc(sizeof(camopts));
  memset(camera->opts, 0, sizeof(camopts));

  _ods_init_defaults(camera);
  _ods_init_env(camera);
  _ods_init_args(camera, argc, argv);

#if DEBUG
  _dump_opts(camera);
#endif

  if (opts->opt_serial)
    {
      odsOpenNSS(camera, camera->opts->device, camera->opts->rate);      
    }
  else if (opts->opt_usb)
    {
      fprintf(stderr, "%s: *** USB not yet supported.\n", argv[0]);
      exit(1);
    }
  else if (opts->opt_irda)
    {
      fprintf(stderr, "%s: *** IrDA not yet supported.\n", argv[0]);
      exit(1);
    }
  
  return camera;
}

void ODSFinalize(ODSCamera camera)
{
  if (camera == 0)
    {
      return;
    }

  if (camera->opts != 0)
    {
      camopts *opts = camera->opts;

      if (opts->opt_serial)
	{
	  odsCloseNSS(camera);
	}
      else if (opts->opt_usb)
	{
	  /* Perform any shutdown required on the USB port */
	}
      else if (opts->opt_irda)
	{
	  /* Perform any shutdown required on the IrDA port */
	}

      strcdup(opts->device, 0);
      strcdup(opts->devtype, 0);
      strcdup(opts->bitrate, 0);

      free(camera->opts);
      camera->opts = 0;
    }

  odsDestroyCamera(camera);
}
