/* Open Digita Services  --  Camera Device Protocol definitions.

   Copyright (C) 1998 James C. Thompson

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc., 59
   Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/*--------------------------------------------------------------------------
 * System include files
 */ 

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

/*--------------------------------------------------------------------------
 * Local include files
 */ 

#include "cameraP.h"
#include "poll-mode-rec.h"
#include "misc.h"

/* Start Editing Here: */

static void fill_bytes_rec(ODSCamera camera);

ODSResult begin_rec(ODSCamera camera,
		     unsigned long *length)
{
  unsigned long unused;
  short command;

  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_prereceive)
    {
      return kODSModeError;
    }

  /* Advance the mode */
  camera->mode = ods_receive;

  /* Make sure we have a buffer to use. */
  if (camera->buffer == 0)
    {
      camera->buffer = malloc(camera->host_size);
      camera->bufsize = camera->host_size;
    }
  else if (camera->bufsize < camera->host_size)
    {
      camera->buffer = realloc(camera->buffer, camera->host_size);
      camera->bufsize = camera->host_size;
    }

  camera->bufused = 0;
  camera->bufbytes = 0;

  /* Set up false information about the message length and bytes-sent
     so we can properly receive the true length. */
  camera->message_length = 4;
  camera->message_sent = 0;
  
  /* Receive the message length */
  rec_uint(camera, length);

  /* Set up the true message length. */
  camera->message_length = *length;
  camera->message_sent = 0;

  /* Receive the remaining parameters in the message header. */
  rec_uint(camera, &unused);
  rec_ushort(camera, &command);
  rec_ushort(camera, &(camera->result));

  /* Check the command to ensure that it matches the last sent
     command.  If not, something is amiss. */
  if ((~command) != ((unsigned short) camera->command))
    {
      fprintf(stderr, "Command mismatch: sent %02hx, got result for %02hx\n",
	      (unsigned short) (camera->command), ~command);
    }

  return kODSNoErr;
}

ODSResult rec_ushort(ODSCamera camera, short *s)
{
  ODSResult rc;

  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_receive)
    {
      return kODSModeError;
    }

  rc = rec_bytes(camera, s, sizeof(*s));
  swap_short(*s);
  return rc;
}

ODSResult rec_uint(ODSCamera camera, UInteger *u)
{
  ODSResult rc;

  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_receive)
    {
      return kODSModeError;
    }

  rc = rec_bytes(camera, u, sizeof(*u));
  swap_long(*u);
  return rc;
}

ODSResult rec_sint(ODSCamera camera, SInteger *s)
{
  ODSResult rc;

  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_receive)
    {
      return kODSModeError;
    }

  rc = rec_bytes(camera, s, sizeof(*s));
  swap_long(*s);
  return rc;
}

ODSResult rec_fixed(ODSCamera camera, Fixed *f)
{
  ODSResult rc;

  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_receive)
    {
      return kODSModeError;
    }

  rc = rec_bytes(camera, f, sizeof(*f));
  swap_long(*f);
  return rc;
}

ODSResult rec_boolean(ODSCamera camera, Boolean *b)
{
  ODSResult rc;

  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_receive)
    {
      return kODSModeError;
    }

  rc = rec_bytes(camera, b, sizeof(*b));
  swap_long(*b);
  return rc;
}

ODSResult rec_bitflags(ODSCamera camera, BitFlags *f)
{
  ODSResult rc;

  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_receive)
    {
      return kODSModeError;
    }

  rc = rec_bytes(camera, f, sizeof(*f));
  swap_long(*f);
  return rc;
}

ODSResult rec_pname(ODSCamera camera, PName *p)
{
  ODSResult rc;

  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_receive)
    {
      return kODSModeError;
    }

  rc = rec_bytes(camera, p, sizeof(p));
  swap_long(*p);
  return rc;
}

ODSResult rec_string(ODSCamera camera, String *s)
{
  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_receive)
    {
      return kODSModeError;
    }

  return rec_bytes(camera, s, sizeof(*s));
}

ODSResult rec_dosname(ODSCamera camera, DOSName *d)
{
  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_receive)
    {
      return kODSModeError;
    }

  return rec_bytes(camera, d, sizeof(*d));
}

ODSResult rec_bytes(ODSCamera camera, void *addr, size_t len)
{
  unsigned long unused;

  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_receive)
    {
      return kODSModeError;
    }

  unused = camera->bufbytes - camera->bufused;

  /* If we need more bytes than are in the buffer, we'll have to read some. */
  while (len > unused)
    {
      /* If there are *some* bytes in the buffer, copy them over */
      if (unused > 0)
	{
	  if (addr != 0)
	    {
	      memcpy(addr, camera->buffer + camera->bufused, unused);
	      addr += unused;
	    }
	  camera->message_sent += unused;
	  len -= unused;
	}

      if (camera->message_sent >= camera->message_length)
	{
	  if (addr != 0)
	    {
	      memset(addr, 0, len);
	    }
	  return kODSBufferError;
	}

      /* Refill the buffer with bytes from the camera; this will reset
	 bufused. */
      fill_bytes_rec(camera);

      unused = camera->bufbytes - camera->bufused;
    }

  if (len > 0)
    {
      if (addr != 0)
	{
	  memcpy(addr, camera->buffer + camera->bufused, len);
	}
      camera->bufused += len;
      camera->message_sent += len;
    }

  return kODSNoErr;
}

ODSResult end_rec(ODSCamera camera)
{
  unsigned short endmark;

  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_receive)
    {
      return kODSModeError;
    }

/*   printf("end_rec: message length %ld, sent so far %ld\n", */
/* 	 camera->message_length, camera->message_sent); */

  /* If any bytes remain unread in the message, read them now.  The
     necessity for doing this was discovered in receiving thumbnails.
     See the "Oddity" note in doc/ERRATA. */
  if (camera->message_sent < camera->message_length)
    {
      rec_bytes(camera, 0, camera->message_length - camera->message_sent);
    }

  read_fully(camera->fd, &endmark, 2);

  camera->mode = ods_presend;

  return kODSNoErr;
}

static void poll_and_reply(ODSCamera camera, const ODSPoll *poll, 
			   ODSPollAck *poll_ack)
{
  unsigned short s;
  read_fully(camera->fd, &s, sizeof(s));
  swap_short(s);
  *((unsigned short *) poll) = s;
  
  poll_ack->unused = 0;
  poll_ack->ack = 1;		/* Blindly accept */
  poll_ack->nak = 0;
  
  s = *((unsigned short *) poll_ack);
  swap_short(s);
  write_fully(camera->fd, &s, sizeof(s));
}

static void fill_bytes_rec(ODSCamera camera)
{
  ODSPoll poll;
  ODSPollAck poll_ack;

  poll_and_reply(camera, &poll, &poll_ack);

  if (poll.length > camera->bufsize)
    {
      camera->bufsize = poll.length;
      camera->buffer = realloc(camera->buffer, poll.length);
    }

  read_fully(camera->fd, camera->buffer, poll.length);
  camera->bufbytes = poll.length;
  camera->bufused = 0;
}
