/* Open Digita Services  --  Camera Device Protocol definitions.

   Copyright (C) 1998 James C. Thompson

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc., 59
   Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/*--------------------------------------------------------------------------
 * System include files
 */ 

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

/*--------------------------------------------------------------------------
 * Local include files
 */ 

#include "cameraP.h"
#include "poll-mode-send.h"
#include "misc.h"

/* Start Editing Here: */

static void flush_bytes_send(ODSCamera camera);

ODSResult begin_send(ODSCamera camera,
		     unsigned long length,
		     ods_command command)
{
  /* Check the passed arguments. */
  if (camera == 0)
    {
      return kODSParamError;
    }

  /* Ensure that we're in the right mode. */
  if (camera->mode != ods_presend)
    {
      return camera->error = kODSModeError; /* Assignment */
    }

  /* Advance the mode. */
  camera->mode = ods_send;

  /* Make sure we have a buffer to use. */
  if (camera->buffer == 0)
    {
      camera->buffer = malloc(camera->device_size);
      camera->bufsize = camera->device_size;
    }
  else if (camera->bufsize < camera->device_size)
    {
      camera->buffer = realloc(camera->buffer, camera->device_size);
      camera->bufsize = camera->device_size;
    }

  /* Set up the variables that track the message size, number of bytes 
     sent, etc. */
  camera->message_length = length + 12; /* Add 12 to account for the
					   words we send. */
  camera->message_sent = 0;
  camera->bufused = 0;

  /* Send the size, version, and reserved bytes. */
  send_uint(camera, camera->message_length - 4);
  send_uint(camera, 0);

  /* Send the command and (unused) result. */
  camera->command = command;
  send_ushort(camera, (unsigned short) command);
  send_ushort(camera, 0);

  /* This returns an error if any of the send_ functions encountered
     an error. */
  return camera->error;
}

ODSResult send_ushort(ODSCamera camera, short s)
{
  /* Check the passed arguments. */
  if (camera == 0)
    {
      return kODSParamError;
    }

  /* Ensure that we're in the right mode. */
  if (camera->mode != ods_send)
    {
      return camera->error = kODSModeError; /* Assignment */
    }

  swap_short(s);
  return send_bytes(camera, &s, sizeof(s));
}

ODSResult send_uint(ODSCamera camera, UInteger u)
{
  /* Check the passed arguments. */
  if (camera == 0)
    {
      return kODSParamError;
    }

  /* Ensure that we're in the right mode. */
  if (camera->mode != ods_send)
    {
      return camera->error = kODSModeError; /* Assignment */
    }

  swap_long(u);
  return send_bytes(camera, &u, sizeof(u));
}

ODSResult send_sint(ODSCamera camera, SInteger s)
{
  /* Check the passed arguments. */
  if (camera == 0)
    {
      return kODSParamError;
    }

  /* Ensure that we're in the right mode. */
  if (camera->mode != ods_send)
    {
      return camera->error = kODSModeError; /* Assignment */
    }

  swap_long(s);
  return send_bytes(camera, &s, sizeof(s));
}

ODSResult send_fixed(ODSCamera camera, Fixed f)
{
  /* Check the passed arguments. */
  if (camera == 0)
    {
      return kODSParamError;
    }

  /* Ensure that we're in the right mode. */
  if (camera->mode != ods_send)
    {
      return camera->error = kODSModeError; /* Assignment */
    }

  swap_long(f);
  return send_bytes(camera, &f, sizeof(f));
}

ODSResult send_boolean(ODSCamera camera, Boolean b)
{
  /* Check the passed arguments. */
  if (camera == 0)
    {
      return kODSParamError;
    }

  /* Ensure that we're in the right mode. */
  if (camera->mode != ods_send)
    {
      return camera->error = kODSModeError; /* Assignment */
    }

  swap_long(b);
  return send_bytes(camera, &b, sizeof(b));
}

ODSResult send_bitflags(ODSCamera camera, BitFlags f)
{
  /* Check the passed arguments. */
  if (camera == 0)
    {
      return kODSParamError;
    }

  /* Ensure that we're in the right mode. */
  if (camera->mode != ods_send)
    {
      return camera->error = kODSModeError; /* Assignment */
    }

  swap_long(f);
  return send_bytes(camera, &f, sizeof(f));
}

ODSResult send_pname(ODSCamera camera, PName p)
{
  /* Check the passed arguments. */
  if (camera == 0)
    {
      return kODSParamError;
    }

  /* Ensure that we're in the right mode. */
  if (camera->mode != ods_send)
    {
      return camera->error = kODSModeError; /* Assignment */
    }

  swap_long(p);
  return send_bytes(camera, &p, sizeof(p));
}

ODSResult send_string(ODSCamera camera, const String *s)
{
  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_send)
    {
      return kODSModeError;
    }

  return send_bytes(camera, s, sizeof(*s));
}

ODSResult send_dosname(ODSCamera camera, const DOSName *d)
{
  if (camera == 0)
    {
      return kODSParamError;
    }

  if (camera->mode != ods_send)
    {
      return kODSModeError;
    }

  return send_bytes(camera, d, sizeof(*d));
}

ODSResult send_bytes(ODSCamera camera, const void *addr, size_t len)
{
  /* Check the passed arguments. */
  if (camera == 0)
    {
      return kODSParamError;
    }

  /* Ensure that we're in the right mode. */
  if (camera->mode != ods_send)
    {
      return camera->error = kODSModeError; /* Assignment */
    }

  /* If this would overflow the buffer, then we'll have to copy part,
     flush, and then copy the rest. */

  if ((camera->bufused + len) >= camera->device_size)
    {
      int partial = 
	(camera->bufused + len) - camera->device_size; /* Number of bytes
							  to copy before
							  flushing. */
      /* Copy as many bytes as will fit... */
      memcpy(camera->buffer + camera->bufused, addr, partial);
      camera->bufused += partial;

      /* And send the buffer to the camera. */
      flush_bytes_send(camera);

      camera->bufused = 0;

      /* Now copy what remains, if anything. */
      len -= partial;
      if (len > 0)
	{
	  memcpy(camera->buffer, addr, len);
	}
    }
  else
    {
      char *buffer = (char *) (camera->buffer);
      buffer += (camera->bufused);
      memcpy(buffer, addr, len);
      camera->bufused += len;
    }

  return kODSNoErr;
}

ODSResult end_send(ODSCamera camera)
{
  /* Check the passed arguments. */
  if (camera == 0)
    {
      return kODSParamError;
    }

  /* Ensure that we're in the right mode. */
  if (camera->mode != ods_send)
    {
      return camera->error = kODSModeError; /* Assignment */
    }

  flush_bytes_send(camera);

  camera->mode = ods_prereceive;

  return kODSNoErr;
}

static void poll_and_wait(ODSCamera camera, const ODSPoll *poll, 
			  ODSPollAck *poll_ack)
{
  unsigned short s = *((unsigned short *) poll);
  s = swap_short(s);
  write_fully(camera->fd, &s, sizeof(s));

  read_fully(camera->fd, &s, sizeof(s));
  s = swap_short(s);
  *((unsigned short *) poll_ack) = s;
}

static void flush_bytes_send(ODSCamera camera)
{
  ODSPoll poll;
  ODSPollAck poll_ack;
  short endmark = 0;

  /* If no bytes have yet been sent, then this is the first packet --
     do the poll-and-wait before sending any bytes. */
  if (camera->message_sent == 0)
    {
      poll.poll = 1;
      poll.cmd_flag = 1;	/* Command */
      poll.bob_flag = 1;	/* Beginning of buffer */
      poll.eob_flag =		/* End of buffer? */
	(camera->bufused == camera->message_length);
      poll.length = camera->bufused;

      poll_and_wait(camera, &poll, &poll_ack);

      if (poll_ack.nak != 0)
	{
	  /* What we really need here is a way to reduce the buffer
	     size and retry. */
	  fprintf(stderr, "*** initial poll NAK.\n");
	}
    }

  camera->message_sent += camera->bufused;

  /* Figure out how many bytes we will send in the next polled
     block. */
  if (camera->message_sent > camera->message_length)
    {
      fprintf(stderr, "*** warning: attempt to send too many bytes.\n");
    }

  /* Prepare the poll to follow this message. */
  poll.poll = 0;		/* ignored */
  poll.cmd_flag = 0;		/* ignored */
  poll.bob_flag = 0;		/* ignored */
  poll.eob_flag = 0;		/* ignored */
  poll.length = 0;

  usleep(2000);

  write_fully(camera->fd, camera->buffer, camera->bufused);
  write_fully(camera->fd, &endmark, 2);

  if (poll_ack.nak != 0)
    {
      /* What we really need here is a way to reduce the buffer
	 size and retry. */
      fprintf(stderr, "*** poll NAK.\n");
    }
}
