/* Open Digita Services  --  Camera Device Protocol definitions.

   Copyright (C) 1998 James C. Thompson

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2 of the License, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc., 59
   Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

/*--------------------------------------------------------------------------
 * System include files
 */ 

#include <stdio.h>
#include <unistd.h>

/*--------------------------------------------------------------------------
 * Local include files
 */ 

#include "cdpP.h"
#include "wire.h"
#include "misc.h"
#include "poll-mode-rec.h"
#include "poll-mode-send.h"

/* Start editing here: */

int read_beacon(int fd, ODSBeacon *beacon)
{
  int c = read_fully(fd, beacon, sizeof_beacon);
    
#if defined(DEBUG)
  dump("read beacon", c, beacon);
#endif

  return c == sizeof_beacon;
}

int check_beacon(ODSBeacon *beacon)
{
  return verify_checksum(beacon, sizeof_beacon);
}

void swap_beacon(ODSBeacon *beacon)
{
  swap_short(beacon->intro);
  swap_short(beacon->vendor_id);
  swap_short(beacon->device_id);
}

void dump_beacon(ODSBeacon *beacon)
{
  printf("Beacon: intro = 0x%04hx, vendor = 0x%04hx, device = 0x%04hx\n",
	 beacon->intro, beacon->vendor_id, beacon->device_id);
}

int write_beacon_ack(int fd, ODSBeaconAck *beacon_ack)
{
  int c = write_fully(fd, beacon_ack, sizeof_beacon_ack);

#if defined(DEBUG)
  dump("wrote beacon ack", c, beacon_ack);
#endif

  return c == sizeof_beacon_ack;
}

void swap_beacon_ack(ODSBeaconAck *beacon_ack)
{
  swap_short(beacon_ack->intro);
  swap_long(beacon_ack->data_speed);
  swap_short(beacon_ack->device_frame_size);
  swap_short(beacon_ack->host_frame_size);
}

void prepare_beacon_ack(ODSBeaconAck *beacon_ack)
{
  set_checksum(beacon_ack, sizeof_beacon_ack);
}

void dump_beacon_ack(ODSBeaconAck *beacon_ack)
{
  printf("Beacon Ack: intro = %04hx, interface type = %02x\n",
	 beacon_ack->intro, (int) (beacon_ack->interface_type));
  printf("    pod receive mode = %d, host receive mode = %d\n",
	 beacon_ack->cf_pod_receive_mode, beacon_ack->cf_host_receive_mode);
  printf("    data speed = %ld, device frame size = %d, host frame size = %d\n",
	 beacon_ack->data_speed, beacon_ack->device_frame_size,
	 beacon_ack->host_frame_size);
}

int read_beacon_comp(int fd, ODSBeaconComp *beacon_comp)
{
  int c = read_fully(fd, beacon_comp, sizeof_beacon_comp);

#if defined(DEBUG)
  dump("read beacon comp", c, beacon_comp);
#endif

  return c == sizeof_beacon_comp;
}

void swap_beacon_comp(ODSBeaconComp *beacon_comp)
{
  swap_long(beacon_comp->data_speed);
  swap_short(beacon_comp->device_frame_size);
  swap_short(beacon_comp->host_frame_size);
}

void dump_beacon_comp(ODSBeaconComp *beacon_comp)
{
  printf("Beacon Comp: result = %d\n", beacon_comp->result);
  printf("    data speed = %ld, device frame size = %d, host frame size = %d\n",
	 beacon_comp->data_speed, beacon_comp->device_frame_size,
	 beacon_comp->host_frame_size);
}

ODSResult read_ptvs(ODSCamera camera, PNameTypeValueStruct *ptvs)
{
  rec_pname(camera, &(ptvs->Name));
  rec_uint(camera, (UInteger *) &(ptvs->Type));
  rec_bytes(camera, &(ptvs->Value), sizeof(PNameValue));

  switch(ptvs->Type)
    {
    case kUInteger:
      swap_long(ptvs->Value.UInt);
      break;
      
    case kSInteger:
      swap_long(ptvs->Value.Int);
      break;

    case kFixed:
      swap_long(ptvs->Value.fFixed);
      break;
      
    case kBoolean:
      swap_long(ptvs->Value.fBoolean);
      break;

    case kBitFlags:
      swap_long(ptvs->Value.fBitFlags);
      break;

    case kPName:
      swap_long(ptvs->Value.fPName);
      break;

    case kDOSName:
    case kString:
    case kEnumList:
      break;
    }

  return kODSNoErr;
}

void dump_ptvs(PNameTypeValueStruct *ptvs)
{
  char name[5];

  unsigned long pname = ptvs->Name;
  swap_long(pname);
  memcpy(name, (const char *) &pname, 4);
  name[4] = '\0';

  printf("%s: ", name);

  switch(ptvs->Type)
    {
    case kUInteger:
      printf("%08lx\n", ptvs->Value.UInt);
      break;
      
    case kSInteger:
      printf("%08lx\n", ptvs->Value.Int);
      break;

    case kFixed:
      printf("%08lx\n", ptvs->Value.fFixed);
      break;
      
    case kBoolean:
      printf("%s\n", ptvs->Value.fFixed ? "true" : "false");
      break;

    case kBitFlags:
      printf("%08lx\n", ptvs->Value.fBitFlags);
      break;

    case kPName:
      pname = ptvs->Value.fPName;
      swap_long(pname);
      memcpy(name, (const char *) &pname, 4);
      name[5] = '\0';
      printf("\"%s\"\n", name);
      break;

    case kDOSName:
      printf("\"%s\"\n", ptvs->Value.fDOSName);
      break;

    case kString:
      printf("\"%s\"\n", ptvs->Value.fString);
      break;

    case kEnumList:
      break;
    }
}

ODSResult write_filename(ODSCamera camera, const FileNameStruct *filename)
{
  send_uint(camera, filename->DriveNo);
  send_string(camera, &(filename->PathName));
  send_dosname(camera, &(filename->DOSName));

  return kODSNoErr;
}

ODSResult read_fileitem(ODSCamera camera, ResFileItem *fileitem)
{
  rec_sint(camera, &(fileitem->DriveNo));
  rec_string(camera, &(fileitem->PathName));
  rec_dosname(camera, &(fileitem->DOSName));
  rec_sint(camera, &(fileitem->FileLength));
  rec_bitflags(camera, &(fileitem->FileStatus));

  return kODSNoErr;
}

ODSResult write_partialtag(ODSCamera camera, const PartialTag *partialtag)
{
  send_uint(camera, partialtag->Offset);
  send_uint(camera, partialtag->Length);
  send_uint(camera, partialtag->Filesize);

  return kODSNoErr;
}

ODSResult read_partialtag(ODSCamera camera, PartialTag *partialtag)
{
  rec_uint(camera, &(partialtag->Offset));
  rec_uint(camera, &(partialtag->Length));
  rec_uint(camera, &(partialtag->Filesize));

  return kODSNoErr;
}
